unit PCapWorks;

interface
uses
  PCap,
  bpf,
  Packet32,
  Windows,
  Messages,
  Classes;

const
  WM_TIMER_EVENT = WM_USER + $200;
  MAX_PCAP_BUF = 512000;
  MAX_UDP_BUF = 8192;

type
  PIP_HDR = ^IP_UDP_HDR;
  IP_UDP_HDR = packed record
    Destination:    array[0..5] of UCHAR;
    Source:         array[0..5] of UCHAR;
    Protocol:       array[0..1] of UCHAR;
    Data:           array[0..0] of UCHAR;
    Verlen:         UCHAR;
    Service:        UCHAR;
    Length:         array[0..1] of UCHAR;
    Ident:          array[0..1] of UCHAR;
    Flagoff:        array[0..1] of UCHAR;
    TimeLive:       UCHAR;
    IpProtocol:     UCHAR;
    Checksum:       array[0..1] of UCHAR;
    IpSource:       array[0..3] of UCHAR;
    IpDestination:  array[0..3] of UCHAR;
    IpData:         array[0..0] of UCHAR;
    UdpSource:      array[0..1] of UCHAR;
    UdpDestination: array[0..1] of UCHAR;
    UdpLength:      array[0..1] of UCHAR;
    UdpCheckSum:    array[0..1] of UCHAR;
    UdpData:        array[0..0] of UCHAR;
  end;

  TRcv_nfo = record
    IpSource:       LongWord;
    IpDest:         LongWord;
    IpProtocol:     Byte;
    UdpSource:      Word;
    UdpDest:        Word;
  end;

  TGTOnPacket = procedure(buf: array of Char; buf_len: Word; rcv_nfo: TRcv_nfo) of object;

  {TGrabThread Object}
  TGrabThread = class(TThread)
  protected
    procedure DoWork;
    procedure Execute; override;
  private
    FGAdapter: PAdapter;
    FGPacket: PPacket;
    FGbp: Pointer;
    FGPort: Word;
    FOnPacket: TGTOnPacket;
    procedure GetPackets(Packet: PPacket; bp: Pointer);
  public
    constructor Create(Adapter: PAdapter; Packet: PPacket; buf: Pointer; Port: Word);
  published
    property OnPacket: TGTOnPacket read FOnPacket write FOnPacket;
  end;

implementation

  {--- TGrabThread -------------------------------------------------------------}

  constructor TGrabThread.Create(Adapter: PAdapter; Packet: PPacket; buf: Pointer; Port: Word);
  begin
    FGAdapter := Adapter;
    FGPacket := Packet;
    FGbp := buf;
    FGPort := Port;
    inherited Create(False);
  end;

  procedure TGrabThread.Execute;
  begin
    while not Terminated do
      DoWork;
  end;

  procedure TGrabThread.DoWork;
  begin
    if PacketReceivePacket(FGAdapter, FGPacket, True) then
      GetPackets(FGPacket, FGbp);
  end;

  procedure TGrabThread.GetPackets(Packet: PPacket; bp: Pointer);
  var
    hdrlen, caplen: LongWord;
    ulBytesReceived: LongWord;
    BufPtr: Pointer;
    i: Word;
    _PChar: PChar;
    Lport: Word;
    buf: array[0..MAX_UDP_BUF - 1] of Char;
    nfo: TRcv_nfo;
  begin
    ulBytesReceived := Packet.ulBytesReceived;
    BufPtr := Ptr(LongWord(bp) + ulBytesReceived); //get pointer to end of received buffer
    while (LongWord(bp) < LongWord(BufPtr)) do
    begin
      FillChar(buf, SizeOf(buf), 0);
      caplen := Pbpf_hdr(bp).bh_caplen;
      hdrlen := Pbpf_hdr(bp).bh_hdrlen;
      _PChar := Ptr(LongWord(bp) + hdrlen);
      LongWord(bp) := LongWord(bp) + BPF_WORDALIGN(caplen + hdrlen);      
      LPort := PIP_HDR(_PChar).UdpSource[0] shl 8 + PIP_HDR(_PChar).UdpSource[1];
      if Lport = FGPort then
      begin
        for i := 42 to caplen - 1 do //skip record size
          if High(buf) > i - 42 then
            buf[i - 42] := Char(_PChar[i]);
        nfo.IpSource   := PIP_HDR(_PChar).IpSource[0] shl 16 + PIP_HDR(_PChar).IpSource[1] shl 8 + PIP_HDR(_PChar).IpSource[2];
        nfo.IpDest     := PIP_HDR(_PChar).IpDestination[0] shl 16 + PIP_HDR(_PChar).IpDestination[1] shl 8 + PIP_HDR(_PChar).IpDestination[2];
        nfo.IpProtocol := PIP_HDR(_PChar).IpProtocol;
        nfo.UdpSource  := PIP_HDR(_PChar).UdpSource[0] shl 8 + PIP_HDR(_PChar).UdpSource[1];
        nfo.UdpDest    := PIP_HDR(_PChar).UdpDestination[0] shl 8 + PIP_HDR(_PChar).UdpDestination[1];
        if Assigned(OnPacket) then
          FOnPacket(buf, caplen - 42, nfo);
        end;
      end;
    end;
end.
