unit Window;

interface
uses
  PCapWorks;

type
  Tpvt_msg = record
    id: String[9];
    FromUser: String;
    ToUser: String;
    Msg: String;
    Multi: Boolean;
  end;

  Tpvt_chat = record
    id: String[9];
    ctype: Char;
    FromUser: String;
    ToUser: String;
    gender: Char;
    Msg: String;
  end;


  Tpvt_conf = record
    id: String[9];
    status: Char;
    FromUser: String;
    ToUser: String;
    gender: Char;
    currentAA: String;
  end;

  Tpkt_proc = class
  public
    procedure onpkt(buf: array of Char; buf_len: Word; rcv_nfo: TRcv_nfo);
  end;

function WndProc(hwnd, wmsg, wParam, lParam: Integer): Integer; stdcall;

implementation
uses
  PCap,
  Classes,
  Packet32,
  Messages,
  Windows,
  SysUtils;

var
  Fadapters: TStringList;
  FAdapter: PAdapter;
  FPacket: PPacket;
  PktProc: Tpkt_proc;
  buf: array[0..512000] of Char;
  GThread: TGrabThread;

function WndProc(hwnd, wmsg, wParam, lParam: Integer): Integer; stdcall;
begin
  Result := 0;
  Case wmsg Of
    WM_CREATE:
      begin
        PktProc := Tpkt_proc.Create;
        Fadapters := TStringList.Create;
        Fadapters.commatext := Pcap.pcap_GetAdapternames(',');
        FAdapter := PacketOpenAdapter(PChar(Fadapters.Strings[0]));
        if (FAdapter = nil) or (FAdapter.hFile = INVALID_HANDLE_VALUE) then
          Exit;
        PacketSetHwFilter(FAdapter, $0008); //accept only broadcast packets
        PacketSetBuff(FAdapter, 512000);
        PacketSetReadTimeout(FAdapter, 10000);
        FPacket := PacketAllocatePacket;
        GThread := TGrabThread.Create(FAdapter, FPacket, @buf, 8167);
        GThread.OnPacket := PktProc.onpkt;
        PacketInitPacket(FPacket, @buf, 512000);
      end;
    WM_DESTROY:
      begin
        if GThread <> nil then
        begin
          GThread.Suspend;
          GThread.Terminate;
          GThread.free;
        end;
        { destroy tray icon }
        {Shell_NotifyIcon(NIM_DELETE, @_IconData);}
        PostQuitMessage(0);
      end;
    else
      Result := DefWindowProc(hwnd, wmsg, wparam, lparam);
  end;
end;

function ab(var a: array of Char; len: Word): String;
var
  i: Integer;
  b: String;
begin
  b := '';
  for i := Low(a) to len do
    if a[i] <> #0 then b := b + a[i] else
      b := b + '#0';
  Result := b;
end;

function npos(var a: array of Char; skip: Word = 0): Word;
var
  skp, i: Word;
begin
  Result := 0; skp := 0;
  for i := Low(a) to High(a) do
    if a[i] = #0 then
    begin
      if skp = skip then
      begin
        Result := i;
        Break;
      end;
      Inc(skp);
    end;
end;

function pcpy(var a: array of Char; index: Word; count: Word): String;
var
  b: String;
  i: Word;
begin
  b := '';
  if count > 0 then
    for i := index to index + count do
      b := b + a[i];
  Result := b;
end;

function DirectoryExists(const Name: string): Boolean;
var
  Code: Integer;
begin
  Code := GetFileAttributes(PChar(Name));
  Result := (Code <> -1) and (FILE_ATTRIBUTE_DIRECTORY and Code <> 0);
end;

procedure DoWrite(FromUser: String; ToUser: String; msg: String; fsuffix: String = '');
function ReChar(S: String): String;
var
  i: Word;
begin
  Result := '';
  for i := 1 to Length(S) do
    if (S[i] in ['A'..'z']) or
       (S[i] in ['0'..'9']) or
       (S[i] in [''..'']) or
       (S[i] in ['!','#','@','$','%','^','&','*','-','+','?']) then
         Result := Result + S[i]
    else
      Result := Result + '_';
end;

var
  F: TextFile;
begin
  FromUser := ReChar(FromUser);  //remove non safe chars from file-names
  ToUser := ReChar(ToUser);
  if not DirectoryExists(FromUser) then
    if not CreateDir(FromUser) then
      Exit;
  if not FileExists(FromUser + '\' + ToUser + fsuffix) then
  begin
    Assign(F, FromUser + '\' + ToUser + fsuffix);
    {$I-}
    ReWrite(F);
    if IOResult <> 0 then
      Exit;
    {$I+}
    CloseFile(F);
  end;
  Assign(F, FromUser + '\' + ToUser + fsuffix);
  {$I-}
  Append(F);
  if IOResult <> 0 then
    Exit;
  Writeln(F, msg);
  {$I+}
  CloseFile(F);
end;

procedure LogMsg(pvt_msg: Tpvt_msg);
var
  tstr, m: String;
begin
  m := '';
  if pvt_msg.Multi then
    m := ' (M)';
  tstr := pvt_msg.FromUser + '>' + pvt_msg.ToUser + ' (out)' + m + #13#10 +
          DateToStr(Date) + ' ' + TimeToStr(Time) + #13#10 +
          pvt_msg.Msg + #13#10;
  DoWrite(pvt_msg.FromUser, pvt_msg.ToUser, tstr);
  tstr := pvt_msg.ToUser + '<' + pvt_msg.FromUser + ' (in)' + m + #13#10 +
          DateToStr(Date) + ' ' + TimeToStr(Time) + #13#10 +
          pvt_msg.Msg + #13#10;
  DoWrite(pvt_msg.ToUser, pvt_msg.FromUser, tstr);
end;

procedure LogConf(pvt_conf: Tpvt_conf);
var
  tstr: String;
begin
  tstr := 'Conf.';
  if pvt_conf.status <> '0' then
  begin
    case pvt_conf.status of
      '1':  tstr := tstr + ' [DND] ';
      '2':  tstr := tstr + ' [Away] ';
      '3':  tstr := tstr + ' [Offline] ';
    end;
    if pvt_conf.currentAA <> '' then
      tstr := tstr + '(' + pvt_conf.currentAA+ ')';
  end;
  tstr := tstr + #13#10;
  DoWrite(pvt_conf.FromUser, pvt_conf.ToUser, tstr);
end;

procedure LogChat(pvt_chat: Tpvt_chat);
var
  tstr: String;
begin
  case pvt_chat.ctype of
    '0': tstr := DateToStr(Date) + ' ' + TimeToStr(Time) + #13#10 +
                 pvt_chat.ToUser + ' joines private chat' + #13#10;
    '1': tstr := DateToStr(Date) + ' ' + TimeToStr(Time) + #13#10 +
                 pvt_chat.ToUser + ' leaves private chat' + #13#10;
    '2': tstr := pvt_chat.FromUser + '>' + pvt_chat.ToUser + ' (out) private message' + #13#10 +
                 DateToStr(Date) + ' ' + TimeToStr(Time) + #13#10 +
                 pvt_chat.Msg + #13#10;
    else
      tstr := '';
  end;
  if tstr <> '' then
    DoWrite(pvt_chat.FromUser, pvt_chat.ToUser, tstr, '_pvt');

  case pvt_chat.ctype of
    '0': tstr := DateToStr(Date) + ' ' + TimeToStr(Time) + #13#10 +
                 pvt_chat.FromUser + ' joines private chat' + #13#10;
    '1': tstr := DateToStr(Date) + ' ' + TimeToStr(Time) + #13#10 +
                 pvt_chat.FromUser + ' leaves private chat' + #13#10;
    '2': tstr := pvt_chat.ToUser + '<' + pvt_chat.FromUser + ' (in) private message' + #13#10 +
                 DateToStr(Date) + ' ' + TimeToStr(Time) + #13#10 +
                 pvt_chat.Msg + #13#10;
    else
      tstr := '';
  end;
  if tstr <> '' then
    DoWrite(pvt_chat.ToUser, pvt_chat.FromUser, tstr, '_pvt');
end;

procedure Tpkt_proc.onpkt(buf: array of Char; buf_len: Word; rcv_nfo: TRcv_nfo);
var
  pvt_msg: Tpvt_msg;
  pvt_conf: Tpvt_conf;
  pvt_chat: Tpvt_chat;
begin
  case buf[10] of
    '6', 'E':
    begin
      pvt_msg.id         := Copy(buf, 1, 9);
      pvt_msg.FromUser   := Copy(buf, 12, npos(buf) - 11);
      pvt_msg.ToUser     := Trim(pcpy(buf, npos(buf) + 1, npos(buf, 1) - npos(buf) - 1));
      pvt_msg.Msg        := Trim(pcpy(buf, npos(buf, 1) + 1, npos(buf, 2) - npos(buf, 1) - 1));
      pvt_msg.Multi      := buf[10] = 'E';
      LogMsg(pvt_msg);
    end;
    '7':
    begin
      pvt_conf.id         := Copy(buf, 1, 9);
      pvt_conf.status     := Copy(buf, 12, 1)[1];
      pvt_conf.FromUser   := Copy(buf, 13, npos(buf) - 12);
      pvt_conf.ToUser     := Trim(pcpy(buf, npos(buf) + 1, npos(buf, 1) - npos(buf) - 1));
      pvt_conf.gender     := pcpy(buf, npos(buf, 1) + 1, 1)[1];
      pvt_conf.CurrentAA  := Trim(pcpy(buf, npos(buf, 1) + 2, npos(buf, 2) - npos(buf, 1) - 1));
      LogConf(pvt_conf);
    end;
    'J':
    begin
      pvt_chat.id         := Copy(buf, 1, 9);
      pvt_chat.ctype      := Copy(buf, 12, 1)[1];
      pvt_chat.FromUser   := Copy(buf, 13, npos(buf) - 12);
      pvt_chat.ToUser     := Trim(pcpy(buf, npos(buf) + 1, npos(buf, 1) - npos(buf) - 1));
      case pvt_chat.ctype of
        '2':  pvt_chat.Msg := Trim(pcpy(buf, npos(buf, 1) + 1, npos(buf, 2) - npos(buf, 1) - 1));
        else
          pvt_chat.gender := pcpy(buf, npos(buf, 1) + 1, 1)[1];
      end;
      LogChat(pvt_chat);
    end;
  end;
end;
end.
